# This file is part of the Falcon repository manager
# Copyright (C) 2005-2007 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import falcon
import falcon.util
import shelve, snack, os, re, sys, time
import cPickle as pickle
from gettext import gettext as _

import django.conf
django.conf.settings.configure(
    DATABASE_ENGINE = 'sqlite3',
    INSTALLED_APPS  = ('falcon',)
)
from django.db import models, connection
from django.core import management

falcon_version = '2.0.0~beta2'
prefix         = '/usr'
falcon_author  = 'Dennis Kaarsemaker <dennis@kaarsemaker.net>'
falcon_url     = 'http://www.kaarsemaker.net/software/falcon'

mcre    = re.compile(r'[a-z0-9]+([-.+][a-z0-9]+)*')

class dummy: pass
bogus_options = dummy()
del dummy
bogus_options.force_yes = False
bogus_options.force_no = False
bogus_options.rootdir = ''

class ConfigurationKey(models.Model):
    key = models.CharField(maxlength=50, unique=True)
    value = models.TextField()
    comment = models.TextField(blank=True)

    def __str__(self):
        return self.value()
    def __repr__(self):
        return "<Configuration Key %s: %s>" % (self.key, self.value)

class Configuration:
    def __init__(self, c_prefix='', options = {}):
        self.__dict__['c_prefix'] = c_prefix
        self.__dict__['values'] = {}
        self.__dict__['questions'] = []

    def __repr__(self):
        return '<Falcon configuration>'

    def __str__(self):
        return "\n".join(["%s: %s" % (x, self.__dict__[x]) for x in self.__dict__ \
                          if x[0] != '_' and x not in ('options', 'values', 'questions')]) + \
               "\n" + "\n".join(["%s: %s" % (x, pickle.loads(self.values[x].value)) for x in self.values])

    def configure(self):
        pass
    
    def set_options(self, options):
        for o in options.__dict__:
            if o[0] == '_':
                continue
            self.__dict__[o] = getattr(options,o)

    def register(self, name, default, question=None, validators=None):
        p_name = self.c_prefix + name
        try:
            key = ConfigurationKey.objects.get(key=p_name)
        except ConfigurationKey.DoesNotExist:
            key = ConfigurationKey(key = p_name, value = pickle.dumps(default))
            key.save()
        self.values[p_name] = key
        self.questions.append((name, question, validators))

    def __getattr__(self, key):
        key = self.c_prefix + key
        if key in self.__dict__:
            return self.__dict__[key]
        else:
            v = self.__dict__['values'].get(key)
            if v:
                return pickle.loads(v.value)

    def __getitem__(self, key): 
        key = self.c_prefix + key
        return pickle.loads(self.values[key].value)

    def __setitem__(self, key, value):
        key = self.c_prefix + key
        key = self.__dict__['values'][key]
        key.value = pickle.dumps(value)
        key.save()

    def __setattr__(self, key, value):
        key = self.c_prefix + key
        if key in self.values:
            key = self.values[key]
            key.value = pickle.dumps(value)
            key.save()
        else:
            self.__dict__[key] = value

class SystemConfiguration(Configuration):
    def set_options(self, options, action):
        Configuration.set_options(self, options)
        # Option checking
        if options.force_yes and options.force_no:
            falcon.util.error(_("Can't force both yes and no"))
        if not options.iso_architecture:
            self.iso_architecture = self.architecture
        try:
            self.isosize = int(self.isosize)
        except:
            error(_("Iso size should be a number"))
        # Check rootdir
        if os.path.islink(os.path.join(os.getenv('HOME'), '.falcon', 'rootdir')):
            self.rootdir = os.readlink(os.path.join(os.getenv('HOME'), '.falcon', 'rootdir')) 
        if options.rootdir:
            self.rootdir = options.rootdir
        self.options = options
        if not self.rootdir:
            falcon.util.error(_("No repository root sepcified"))
        if not os.path.exists(self.rootdir):
            falcon.util.error(_("Repository root %s does not exist") % self.rootdir)
        
        # Chdir into the rootdir to avoid some os.path.join calls....
        os.chdir(self.rootdir)

        # Bail out here if creating an iso
        if action == 'iso':
            return

        # Check configuration database
        if not os.path.exists('.falcon'):
            os.mkdir('.falcon')
        falcon.lock()

        # Configure django
        setattr(django.conf.settings, 'DATABASE_NAME',  os.path.join(self.rootdir,'.falcon','data'))

        # Create databases
        management.disable_termcolors()
        all_models = (falcon.pocket.Pocket, falcon.pocket.Component, falcon.pocket.MetaComponent,
                      falcon.package.SourcePackage, falcon.package.BinaryPackage, ConfigurationKey,
                      falcon.mirror.Mirror)
        all_sql = []
        for model in all_models:
            sql, ref = management._get_sql_model_create(model, all_models)
            all_sql += sql
            sql = management._get_many_to_many_sql_for_model(model)
            all_sql += sql
        cursor = connection.cursor()
        for s in all_sql:
            cursor.execute(s.replace('CREATE TABLE', 'CREATE TABLE IF NOT EXISTS'))

        # Register our variables
        self.register('description', _('Falcon generated repository'), 
                      falcon.util.Entry(_("Please enter a short description of your repository")),
                      [])
        self.register('origin', _('falcon'),
                      falcon.util.Entry(_("Please enter a one-word origin for your repository (e.g. your nickname)")),
                      [falcon.util.validators.nospaces])
        self.register('gpgkey', '',
                      falcon.util.Entry(_("Please enter your 8-character GPG id")),
                      [falcon.util.validators.hex])
        all_architectures = ('i386','amd64','powerpc','sparc','arm','ia64','alpha','hppa','m68k','mips','mipsel','s390')
        self.register('architectures', [conf.architecture],
                      falcon.util.SelectMulti(_("Which repository do you want to support"), all_architectures),
                      [falcon.util.validators.atleastone])
        self.register('webbase', '',
                      falcon.util.Entry(_("What is the webroot of your repository? (Leave blank to disable html pages)")),
                      [falcon.util.validators.url])
        templatedirs = [os.path.join(conf.prefix,'share','falcon','templates'),
                        os.path.join(os.getenv('HOME'), '.falcon', 'templates'),
                        os.path.join('.falcon', 'templates')]
        templates = []
        _templatedirs = {}
        for dir in templatedirs:
            if os.path.exists(dir):
                ts = os.listdir(dir)
                templates += ts
                for template in ts:
                    _templatedirs[template] = dir
        templatedirs = _templatedirs
        self.register('template','default',
                      falcon.util.Select(_("Which template do you want to use?"), templates),
                      [])
        self.register('long_description', '',
                      falcon.util.Text(_("Please enter a description for on the webpage")),
                      [])
        self.register('create_filelist',True,
                      falcon.util.Boolean(_("Do you want to create content listings?")),
                      [])
        self.register('configured', False, None, [])
        self.register('lastinstall', time.time(), None,[])
        self.register('lastexport', time.time(), None,[])
        
        # Load/create pockets
        falcon.pockets = []
        for p in sorted(os.listdir('pool')):
            if os.path.isdir(os.path.join('pool',p)):
                falcon.pockets.append(falcon.pocket.Pocket.objects.get_or_create(name=p)[0])
                falcon.pockets[-1].load_components()
        # Remove pockets from the database that no longer exist
        for p in falcon.pocket.Pocket.objects.all():
            if not os.path.exists(p.poolpath):
                p.delete()

        # Load plugins
        falcon.plugin.load_plugins()

        # More django config (needs database)
        if self.template:
            setattr(django.conf.settings, 'TEMPLATE_DIRS', (os.path.join(templatedirs[self.template],self.template),))

    def set_import_options(self, options):
        Configuration.set_options(self, options)
        if os.path.islink(os.path.join(os.getenv('HOME'), '.falcon', 'rootdir')):
            self.rootdir = os.readlink(os.path.join(os.getenv('HOME'), '.falcon', 'rootdir')) 
        if options.rootdir:
            self.rootdir = options.rootdir

        self.options = options
        if not self.rootdir:
            falcon.util.warning(_("No repository root sepcified"))
        elif not os.path.exists(self.rootdir):
            falcon.util.error(_("Repository root %s does not exist") % self.rootdir)

    def configure(self):
        screen = falcon.screen = snack.SnackScreen()
        screen.drawRootText(1,1, _("Falcon %s interactive configuration editor") % self.falcon_version)
        screen.drawRootText(1,2, "(C)2005-2007 %s" % (self.falcon_author))
        self.options.force_yes = False
        self.options.force_no = False

        GENERAL, POCKETS, MIRRORS, PLUGINS, EXIT = range(5)
        while True:
            form = snack.GridFormHelp(screen, _("Configuration wizard"), _("See /usr/share/doc/falcon for falcon documentation"),1,2)
            t = snack.TextboxReflowed(30, _("Please choose from the menu"))
            l = snack.Listbox(5, returnExit=True)
            l.append(_("General configuration"), GENERAL)
            l.append(_("Pockets and components"), POCKETS)
            l.append(_("Mirrors"), MIRRORS)
            l.append(_("Plugins"), PLUGINS)
            l.append(_("Exit configuration"), EXIT)
            form.add(t,0,0)
            form.add(l,0,1, (0,1,0,0))
            item = form.runOnce()
            if item == 'ESC': item = EXIT
            elif item == 'F12':
                item = l.current()
            else:
                item = item.current()

            if item == EXIT:
                screen.suspend()
                self.configured = True
                return
            elif item == GENERAL:
                current = dict([(x[0], self[x[0]]) for x in self.questions if x[1]])
                while True:
                    form = snack.GridFormHelp(screen, _("General configuration"), _("See /usr/share/doc/falcon for falcon documentation"), 
                                             1, sum([x[1].num_newt_widgets for x in self.questions if x[1] != None]) + 1)
                    form.x = 0
                    form.last = -1
                    for q in self.questions:
                        (name, question, validators) = q
                        if question:
                            question.add_newt_widgets(form, current[name])
                    bb = snack.ButtonBar(screen, ((_('OK'),1),(_('Cancel'),2)),compact=True)
                    form.add(bb, 0, form.last+1,padding=(0,1,0,0))
                    ret = form.runOnce()
                    if ret == 'ESC' or bb.buttonPressed(ret) == 2: # Cancel
                        break
                    # Grab all values
                    for q in self.questions:
                        # Validate!
                        if q[1] and not q[1].needs_external_editor:
                            a = q[1].get_answer()
                        elif q[1] and 'button' in dir(q[1]) and q[1].button == ret:
                            a = q[1].button_action(ret, current[q[0]])
                        else:
                            continue
                        try:
                            for val in q[2]:
                               val(a)
                            current[q[0]] = a
                        except falcon.util.ValidationError:
                            # FIXME: Don't fail silently
                            pass
                    # Was the OK button pressed or a config widget?
                    if not isinstance(ret, snack.CompactButton) or (bb.buttonPressed(ret) == 1):
                        for q in self.questions:
                            if q[1]:
                                self[q[0]] = current[q[0]]
                        break
                                
            elif item == POCKETS:
                while True:
                    form = snack.GridFormHelp(screen, _("Pocket configuration"), _("See /usr/share/doc/falcon for falcon documentation"), 
                                              1, 2)
                    t = snack.TextboxReflowed(30, _("Select a pocket to configure"))
                    l = snack.Listbox(5, returnExit=True)
                    l = snack.Listbox(min(5, len(falcon.pocket.Pocket.objects.all())) + 2, returnExit=True,scroll=len(falcon.pocket.Pocket.objects.all()) > 5)
                    for p in falcon.pockets:
                        l.append(p.name,p)
                    l.append(_("Back to main menu"), False)
                    form.add(t,0,0)
                    form.add(l,0,1, (0,1,0,0))
                    item = form.runOnce()
                    if item == 'ESC': break
                    if item == 'F12': item = l
                    item = item.current()
                    if not item: break

                    # Now configure the pocket
                    questions = [
                     (falcon.util.Entry(_("Version number for this pocket")),None),
                     (falcon.util.Text(_("Description of this pocket")),None)
                    ] + [(falcon.util.Text(_("Description for component %s") % x.name),x) for x in item.components]
                    answers = [item.version, item.description] + [x.description for x in item.components]
                    canceled = False
                    while True:
                        form = snack.GridFormHelp(screen, _("Pocket %s") % item.name, _("See /usr/share/doc/falcon for falcon documentation"),
                                              1, 1 + sum([x[0].num_newt_widgets for x in questions]))
                        form.last = -1
                        form.x = 0
                        for i in range(len(questions)):
                            questions[i][0].add_newt_widgets(form, answers[i])
                        bb = snack.ButtonBar(screen, ((_('OK'),-1),(_('Cancel'),-2)),compact=True)
                        form.add(bb, 0, form.last+1,padding=(0,1,0,0))

                        ret = form.runOnce()
                        if ret == 'ESC' or bb.buttonPressed(ret) == -2: # Cancel
                            canceled = True
                            break
                        for i in range(len(questions)):
                            if not questions[i][0].needs_external_editor:
                                a = questions[i][0].get_answer()
                            elif 'button' in dir(questions[i][0]) and questions[i][0].button == ret:
                                a = questions[i][0].button_action(ret, answers[i])
                            else:
                                continue
                            try:
                                if i == 0:
                                    falcon.util.validators.version(a)
                                answers[i] = a
                            except falcon.util.ValidationError:
                                # FIXME
                                pass
                        if not isinstance(ret, snack.CompactButton) or bb.buttonPressed(ret) == -1:
                            item.version = answers[0]
                            item.description = answers[1]
                            item.save()
                            for i in range(2,len(questions)):
                                questions[i][1].description = answers[i]
                                questions[i][1].save()
                            break
                    if canceled:
                        continue
                    questions2 = []
                    answers2 = []
                    for c in item.metacomponents.all():
                        questions2.append((falcon.util.Text(_("Description for metacomponent %s") % c.name), c))
                        answers2.append(c.description)
                        questions2.append((falcon.util.SelectMulti(_("Components"), sorted([y.name for y in item.realcomponents.all()])),c))
                        answers2.append([x.name for x in c.components.all()])
                    questions2.append((falcon.util.Entry(_("Add new metacomponent")),None))
                    answers2.append('')
                    while True:

                        form = snack.GridFormHelp(screen, _("Pocket %s") % item.name, _("See /usr/share/doc/falcon for falcon documentation"),
                                              1, 2 + sum([x[0].num_newt_widgets for x in questions2]))
                        form.last = 0
                        form.x = 0
                        l = snack.Label(_("Metacomponents"))
                        form.add(l, 0, 0)
                        for i in range(len(questions2)):
                            questions2[i][0].add_newt_widgets(form, answers2[i])
                        bb = snack.ButtonBar(screen, ((_('OK'),-1),(_('Cancel'),-2)),compact=True)
                        form.add(bb, 0, form.last+1,padding=(0,1,0,0))

                        ret = form.runOnce()
                        if ret == 'ESC' or bb.buttonPressed(ret) == -2: # Cancel
                            break
                        for i in range(len(questions2)):
                            if not questions2[i][0].needs_external_editor:
                                a = questions2[i][0].get_answer()
                            elif 'button' in dir(questions2[i][0]) and questions2[i][0].button == ret:
                                a = questions2[i][0].button_action(ret, answers2[i])
                            else:
                                continue
                            try:
                                if i == len(questions2) -1:
                                    falcon.util.validators.nospaces(a)
                                    if a:
                                        falcon.pocket.MetaComponent(pocket=item,name=a).save()
                                        c = falcon.pocket.MetaComponent.objects.get(pocket=item,name=a)
                                        qb = questions2[-1]
                                        questions2 = questions2[:-1]
                                        answers2 = answers2[:-1]
                                        questions2.append((falcon.util.Text(_("Description for metacomponent %s") % c.name), c))
                                        answers2.append(c.description)
                                        questions2.append((falcon.util.SelectMulti(_("Components"), sorted([y.name for y in item.realcomponents.all()])),c))
                                        answers2.append([x.name for x in c.components.all()])
                                        questions2.append(qb)
                                        answers2.append(a)
                                answers2[i] = a
                            except falcon.util.ValidationError:
                                # FIXME
                                pass
                        if (not isinstance(ret, snack.CompactButton) or bb.buttonPressed(ret) == -1) and not answers2[-1]:
                            for i in range(len(questions2) -1):
                                comp = questions2[i][1]
                                if type(answers2[i]) == str:
                                    comp.description = answers2[i]
                                    comp.save()
                                else:
                                    comp.components = []
                                    for a in answers2[i]:
                                        comp.components.add(falcon.pocket.Component.objects.get(name=a,pocket=comp.pocket))
                                    if not len(comp.components.all()):
                                        comp.delete()
                                    else:
                                        comp.save()
                            break
                        answers2[-1] = ''

            elif item == MIRRORS:
                while True:
                    form = snack.GridFormHelp(screen, _("Mirror configuration"), _("See /usr/share/doc/falcon for falcon documentation"), 
                                              1, 4)
                    t = snack.TextboxReflowed(30, _("Select a mirror to configure"))
                    l = snack.Listbox(min(5, len(falcon.mirror.Mirror.objects.all())) + 2, returnExit=True,scroll=len(falcon.mirror.Mirror.objects.all()) > 5)
                    for m in falcon.mirror.Mirror.objects.all():
                        l.append(m.name,m)
                    l.append(_("New mirror"), -1)
                    l.append(_("Back to main menu"), False)
                    form.add(t,0,0)
                    form.add(l,0,1, (0,1,0,0))
                    item = form.runOnce()
                    if item == 'ESC': break
                    if item == 'F12': item = l
                    item = item.current()
                    if not item: break

                    if item == -1:
                        item = falcon.mirror.Mirror()
                        comps = []
                    else:
                        comps = list(item.realcomponents.all()) + list(item.metacomponents.all())
                    questions = [
                     (falcon.util.Entry(_("Name of this mirror")), [falcon.util.validators.shellsafe, falcon.util.validators.nospaces]),
                     (falcon.util.Entry(_("URL of this mirror's root")), []),
                     (falcon.util.Entry(_("Rsync upload path")), [falcon.util.validators.shellsafe]),
                     (falcon.util.Entry(_("Extra rsync arguments")), [falcon.util.validators.shellsafe]),
                     (falcon.util.Entry(_("Who sponsors this mirror")), []),
                     (falcon.util.ComponentTree(_("Which components should be mirrored")),[])
                    ]
                    answers = [
                      item.name, item.webbase, item.rsync, item.rsync_opts, item.sponsor, comps
                    ]
                    form = snack.GridFormHelp(screen, _("Mirror %s") % item.name, _("See /usr/share/doc/falcon for falcon documentation"), 
                                              1, 1 + sum([x[0].num_newt_widgets for x in questions]))
                    form.last = -1
                    form.x = 0
                    for i in range(len(questions)):
                        questions[i][0].add_newt_widgets(form, answers[i])
                    bb = snack.ButtonBar(screen, ((_('OK'),-1),(_('Cancel'),-2)),compact=True)
                    form.add(bb, 0, form.last+1,padding=(0,1,0,0))

                    ret = form.runOnce()
                    if ret == 'ESC' or bb.buttonPressed(ret) == -2: # Cancel
                        continue
                    for i in range(len(questions)):
                        a = questions[i][0].get_answer()
                        try:
                            for v in questions[i][1]:
                                v(a)
                            answers[i] = a
                        except falcon.util.ValidationError:
                            #FIXME
                            pass
                    item.name = answers[0]
                    item.webbase = answers[1]
                    item.rsync = answers[2]
                    item.rsync_opts = answers[3]
                    item.sponsor = answers[4]
                    if not len(item.name):
                        item.delete()
                    else:
                        item.save()
                        item = falcon.mirror.Mirror.objects.get(name=item.name)
                        item.realcomponents = [x for x in answers[5] if type(x) == falcon.pocket.Component]
                        item.metacomponents = [x for x in answers[5] if type(x) == falcon.pocket.MetaComponent]
                        item.save()
            elif item == PLUGINS:
                while True:
                    form = snack.GridFormHelp(screen, _("Configuration wizard"), _("See /usr/share/doc/falcon for falcon documentation"),1,2)
                    t = snack.TextboxReflowed(30, _("Please choose from the menu"))
                    l = snack.Listbox(1 + min(8,len(falcon.plugin.loaded_plugins)), returnExit=True, scroll=len(falcon.plugin.loaded_plugins) > 8)
                    for p in falcon.plugin.loaded_plugins.values():
                        l.append(p.name, p)
                    l.append(_("Back to main menu"), False)
                    form.add(t,0,0)
                    form.add(l,0,1, (0,1,0,0))
                    item = form.runOnce()
                    if item == 'ESC':
                        break
                    elif item == 'F12':
                        item = l.current()
                    else:
                        item = item.current()
                    if not item:
                        break
                    item.configure()

# Initial config
conf = SystemConfiguration()
conf.falcon_version = falcon_version
conf.falcon_author  = falcon_author
conf.falcon_url     = falcon_url
conf.prefix         = prefix
conf.template_api   = (1,0)
conf.plugin_api     = (1,0)
conf.database_abi   = (1,0)
conf.architecture   = falcon.util.run(['dpkg-architecture', '-qDEB_HOST_ARCH']).strip()

