# This file is part of the Falcon repository manager
# Copyright (C) 2005-2007 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import falcon
import sys, os, types, snack
from gettext import gettext as _

loaded_plugins = {}
enabled_plugins = []

def load_plugins():
    # Global plugins
    try:
        import falcon.plugins
    except:
        # Extra import is needed here
        import falcon
        falcon.util.debug_exception()
        falcon.util.error(_("Failed to load global plugins"))

    # Homedir plugins
    if os.path.exists(os.path.join(os.getenv('HOME'), '.falcon', 'plugins','__init__.py')):
        os.path.insert(0,os.path.join(os.getenv('HOME'), '.falcon'))
        try:
            import plugins as homedir_plugins
        except:
            falcon.util.error(_("Failed to load plugins in your .falcon"))
        os.path = os.path[1:]

    # Repo plugins
    if os.path.exists(os.path.join('.falcon', 'plugins','__init__.py')):
        os.path.insert(0,os.path.join('.falcon'))
        try:
            import plugins as repo_plugins
        except:
            falcon.util.error(_("Failed to load plugins in your repositories .falcon"))
        os.path = os.path[1:]

class PluginBase(type):
    def __new__(cls, name, bases, attrs):
        # Only do magic on subclasses of the base plugin
        if not bases or bases == (object,):
            return type.__new__(cls, name, bases, attrs)

        new_class = type.__new__(cls, name, bases, attrs)

        # Add a configuration
        new_class.conf = falcon.config.Configuration(c_prefix = 'plugin_%s_' % name)
        new_class.conf.register('enabled', False,
                                falcon.util.Boolean(_("Enable this plugin")), [])
        new_class.__file__ = sys.modules[new_class.__module__].__file__

        # Register the plugin if the api version matches
        try:
            api = new_class.api
            major, minor = int(api[0]), int(api[1])
        except (AttributeError, ValueError, TypeError):
            falcon.util.warning(_("Plugin %s from %s is broken: it doesn't define an API version") % (name, new_class.__file__))
            falcon.util.debug_exception()
            return new_class

        if major != falcon.conf.plugin_api[0] or minor > falcon.conf.plugin_api[1]:
            falcon.util.warning(_("Plugin %s from %s is broken: it is made for a different version of falcon") % (name, new_class.__file__))
            return new_class
        if minor < falcon.conf.plugin_api[1]:
            falcon.util.warning(_("Plugin %s from %s should be upgraded but will still work") % (name, new_class.__file__))
            
        loaded_plugins[name] = new_class
        if new_class.conf.enabled:
            enabled_plugins.append(new_class())
        return new_class

class FalconPlugin:
    
    __metaclass__ = PluginBase

    # Thes 3 should be overridden in actual plugins
    name = "Plugin Interface - you need to change this in your plugin"
    desc = "Badly maintained plugin"
    api  = (0,0)

    # Configuration hook
    @classmethod
    def configure(cls):
        screen = falcon.screen
        current = dict([(x[0], cls.conf[x[0]]) for x in cls.conf.questions])
        while True:
            form = snack.GridFormHelp(screen, cls.name, _("See /usr/share/doc/falcon for falcon documentation"),
                                      1, 2 + sum([x[1].num_newt_widgets for x in cls.conf.questions if x[1]]))
            form.x = 0
            form.last = 0
            form.add(snack.Label(cls.desc), 0, 0)
            for q in cls.conf.questions:
                if q[1]:
                    q[1].add_newt_widgets(form, current[q[0]])
            bb = snack.ButtonBar(screen, ((_('OK'),1),(_('Cancel'),2)),compact=True)
            form.add(bb, 0, form.last+1,padding=(0,1,0,0))
            ret = form.runOnce()

            if ret == 'ESC' or bb.buttonPressed(ret) == 2: # Cancel
                break

            for q in cls.conf.questions:
                # Validate!
                if q[1] and not q[1].needs_external_editor:
                    a = q[1].get_answer()
                elif q[1] and 'button' in dir(q[1]) and q[1].button == ret:
                    a = q[1].button_action(ret, current[q[0]])
                else:
                    continue
                try:
                    for val in q[2]:
                       val(a)
                    current[q[0]] = a
                except falcon.util.ValidationError:
                    # FIXME: Don't fail silently
                    pass
            # Was the OK button pressed or a config widget?
            if not isinstance(ret, snack.CompactButton) or (bb.buttonPressed(ret) == 1):
                was_enabled = cls.conf.enabled
                for q in cls.conf.questions:
                    if q[1]:
                        cls.conf[q[0]] = current[q[0]]
                if cls.conf.enabled and not was_enabled:
                    # Plugin was just enabled, run its constructor and config
                    enabled_plugins.append(cls())
                else:
                    break

    # Before anything is done, right after argument parsing
    def pre_action(self, action, args): pass
    # After everything is done, before quiting
    def post_action(self, action, args): pass

    # When scanning a component
    def pre_scan(self, component): pass
    def post_scan(self, result, component): pass

    # When exporting a component, metacomponent or pocket
    def pre_export(self, component_or_pocket): pass
    def post_export(self, result, component_or_pocket): pass

    # When adding packages to a component (pocket.py)
    def pre_install(self, component, package): pass
    def post_install(self, result, component, package): pass

    # When syncing to a mirror
    def pre_sync(self, mirror): pass
    def post_sync(self, result, mirror): pass

class DontRun(Exception):
    """Prevent the wrapped command from being run"""
    pass

# Decorator chain for plugin hooks
def wrap_plugin(func):
    def wrapped_func(*args, **kwargs):
        res = run_plugins('pre_' + func.func_name, *args, **kwargs)
        if not res:
            return
        result = func(*args, **kwargs)
        run_plugins('post_' + func.func_name, result, *args, **kwargs)
        return result
    return wrapped_func

def run_plugins(name, *args, **kwargs):
    for p in enabled_plugins[:]:
        if name in p.__class__.__dict__:
            try:
                p.__class__.__dict__[name](p, *args, **kwargs)
            except DontRun, e:
                return False
            except Exception, e:
                falcon.util.warning(_("%s has errored out, it will be disabled") % p.name)
                falcon.util.debug_exception()
                enabled_plugins.remove(p)
    return True
