# This file is part of the Falcon repository manager
# Copyright (C) 2005-2007 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import falcon
from gettext import gettext as _
from email.MIMEText import MIMEText
import smtplib, os

class DebcheckerPlugin(falcon.plugin.FalconPlugin):
    name = _("Package checker")
    desc = _("Run packages thorough linda/lintian before installing thm into the archive")
    api  = (1,0)

    def __init__(self):
        self.conf.register('linda', True, falcon.util.Boolean(_("Run packages through linda")), [])
        self.conf.register('linda_args', '', falcon.util.Entry(_("Arguments for linda")),
                           [falcon.util.validators.shellsafe])
        self.conf.register('lintian', True, falcon.util.Boolean(_("Run packages through lintian")), [])
        self.conf.register('lintian_args', '', falcon.util.Entry(_("Arguments for lintian")),
                           [falcon.util.validators.shellsafe])
        self.conf.register('refuse', False, falcon.util.Boolean(_("Refuse to install packages that fail tests")),[])
        self.conf.register('email', True, falcon.util.Boolean(_("Send linda/lintian output to maintainer")),[])
        # Copied from the mail_maintainer plugin
        self.conf.register('frommail', '', falcon.util.Entry(_("Sender address")),
                           [falcon.util.validators.mailaddress])
        self.conf.register('extramail', '', falcon.util.Entry(_("Extra mailaddresses to send to")),
                           [falcon.util.validators.commaseparated(falcon.util.validators.mailaddress)])
        self.conf.register('whitelist', '', falcon.util.Entry(_("File with whitelist of recipients")),[])
        self.conf.register('server', 'localhost', falcon.util.Entry(_("Hostname of the smtp server to use")), [])

    def pre_install(self, component, package):
        flaws_detected = False
        errors = ''
        for f in [package.filename] + [b.filename for b in package.binaries.all()]:
            f = os.path.join(component.poolpath, f)
            if self.conf.linda:
                try:
                    err = falcon.util.run(['linda'] + self.conf.linda_args.split() + [f])
                    err = err.strip()
                    if err:
                        flaws_detected = True
                        errors += err + '\n'
                except RuntimeError, e:
                    flaws_detected = True
                    errors += e.stdout
            if self.conf.lintian:
                try:
                    err = falcon.util.run(['lintian'] + self.conf.lintian_args.split() + [f])
                    err = err.strip()
                    if err:
                        flaws_detected = True
                        errors += err + '\n'
                except RuntimeError, e:
                    flaws_detected = True
                    errors += e.stdout

        if flaws_detected:
            # Largely copied from mail_maintainer.py
            if self.conf.email and self.conf.whitelist:
                whitelist = falcon.util.readfile(self.conf.whitelist)
                addresses = []
                maintainer = package.controlfields['Maintainer']
                if '<' in maintainer:
                    maintainer = maintainer[maintainer.find('<')+1:maintainer.find('>')]
                if maintainer in whitelist:
                    addresses.append(maintainer)
                for m in self.conf.extramail.split(','):
                    addresses.append(m.strip())
                addresses = set(addresses)
                
                msg = MIMEText(_("Package %s (%s) fails linda/lintian checks:\n%s") % (package.packagename, package.version, errors))
                msg['Subject'] = _("Package %s %s fails linda/lintian checks") % (package.packagename, package.version)
                msg['From'] = '"%s" <%s>' % (falcon.conf.description.replace('"',''), self.conf.frommail)
                msg['To'] = ', '.join(addresses)
                try:
                    s = smtplib.SMTP()
                    if falcon.conf.debug:
                        s.set_debuglevel(1)
                    s.connect(self.conf.server)
                    s.sendmail(self.conf.frommail, addresses, msg.as_string())
                    s.quit()
                except:
                    falcon.util.warning(_("Sending e-mail failed"))
                    falcon.util.debug_exception()
            elif self.conf.email:
                falcon.util.warning(_("No whitelist found for mailaddresses, not sending mail"))

            if self.conf.refuse:
                falcon.util.warning(_("There are errors in the %s package. Not installing") % package.packagename)
                package.delete()
                raise falcon.plugin.DontRun()
