# This file is part of the Falcon repository manager
# Copyright (C) 2005-2007 Dennis Kaarsemaker
# See the file named COPYING in the root of the source tree for license details

import falcon
import os, sys, time, subprocess, snack, urllib2, re, tempfile, traceback
from gettext import gettext as _

def writefile(path, data):
    if os.path.dirname(path) and not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    data = str(data)
    if os.path.exists(path) and data == readfile(path):
        debug(_("Not overwriting unchanged file %s") % path)
        return False
    else:
        debug(_("Writing %d bytes to %s") % (len(data), path))
        fd = open(path, 'w')
        fd.write(str(data))
        fd.close()
        return True
 
def readfile(path, ignore_errors = True):
    debug(_("Reading file %s") % path)
    try:
        fd = open(path, 'r')
        d = fd.read()
        fd.close
        return str(d)
    except:
        debug(_("File %s can't be read") % path)
        if ignore_errors:
            return ''
        raise
    
def run(command, wd=None, buffer=True, outfile=None):
    if wd:
        debug(_("Executing in workdir %s: %s") % (wd, command))
    else:
        debug(_("Executing: %s") % command)
    out = subprocess.PIPE
    err = subprocess.PIPE
    if not buffer:
        out = None
        err = None
    if outfile:
        out = open(outfile,'w')
    process = subprocess.Popen(command, cwd = wd, stdout = out, stderr = err, stdin = subprocess.PIPE)
    stdout, stderr = process.communicate()
    if outfile:
        out.close()
    ret = process.wait()
    if ret != 0:
        ex = RuntimeError(_("%s failed with code %d\n\n%s") % (command, ret, stderr))
        ex.code = ret
        ex.stdout = stdout
        ex.stderr = stderr
        raise ex
    return stdout

def intersect(a1,a2):
    if a1 == ['all']:
        return a2
    if a2 == ['all']:
        return a1
    return [x for x in a1 if x in a2]

output  = lambda x: sys.stdout.write('*  ' + x + "\n")
warning = lambda x: sys.stderr.write('W: ' + x + "\n")

def error(msg):
    print 'E: ' + msg
    if falcon.screen:
        falcon.screen.finish()
    falcon.unlock()
    sys.exit(1)
def debug(msg):
    if 'conf' in dir(falcon) and falcon.conf.verbose:
        print 'D: ' + msg

outfunc = (output, warning, error, debug)

def newer(file1, file2):
    try:
        return os.path.getmtime(file1) > os.path.getmtime(file2)
    except: # Exceptions occur when files don't exist
        return True

# The timer decorator
def timer(func):
    def wrapper(*args, **kwargs):
        if not falcon.conf.timings:
            return func(*args, **kwargs)
        start = time.time()
        res = func(*args, **kwargs)
        stop = time.time()
        delta = stop - start
        falcon.util.output(_("Function %s (args: %s, kwargs: %s) took %.3f seconds") %
                           (func.func_name, str(args), str(kwargs), delta))
        return res
    return wrapper

# Timer stack for non-decorator timers
timerstack = []
def timer_start(text=None):
    if falcon.conf.timings:
        timerstack.append((text,time.time()))

def timer_stop():
    if falcon.conf.timings:
        text, tm = timerstack.pop()
        falcon.util.output(_("%s took %.4f seconds") % (text, time.time() - tm))

# Configuration bits
class Question:
    needs_external_editor = False
    num_newt_widgets = 2
    def __init__(self, question):
        self.question = question

class Entry(Question):
    def add_newt_widgets(self, form, current):
        label = snack.Label(self.question)
        self.entry = snack.Entry(30, str(current), returnExit = True)
        form.add(label, form.x, form.last + 1)
        form.add(self.entry, form.x, form.last + 2)
        form.last += 2

    def get_answer(self):
        return self.entry.value()

class Select(Question):
    def __init__(self, question, selections):
        self.question = question
        self.selections = selections

    def add_newt_widgets(self, form, current):
        label = snack.Label(self.question)
        self.listbox = snack.Listbox(min(5,len(self.selections)),scroll=len(self.selections) > 5,returnExit=True) 
        for s in self.selections:
            self.listbox.append(s,s)
        self.listbox.setCurrent(current)
        form.add(label, form.x, form.last + 1)
        form.add(self.listbox, form.x, form.last + 2)
        form.last += 2

    def get_answer(self):
        return self.listbox.current()

class SelectMulti(Select):
    def add_newt_widgets(self, form, current):
        label = snack.Label(self.question)
        self.tree = snack.CheckboxTree(min(5,len(self.selections)), scroll=len(self.selections) > 5) 
        for s in self.selections:
            self.tree.append(s, selected=(s in current))
        form.add(label, form.x, form.last + 1)
        form.add(self.tree, form.x, form.last + 2)
        form.last += 2

    def get_answer(self):
        return self.tree.getSelection()

class ComponentTree(Question):
    def add_newt_widgets(self, form, current):
        label = snack.Label(self.question)
        self.tree = snack.CheckboxTree(5, scroll=True)
        i = 0
        for p in falcon.pockets:
            self.tree.append(p.name)
            for c in p.components:
                self.tree.addItem(c.name,(i,-1),c,c in current)
            i += 1
        form.add(label, form.x, form.last + 1)
        form.add(self.tree, form.x, form.last + 2)
        form.last += 2
    def get_answer(self):
        return self.tree.getSelection()

class Text(Question):
    needs_external_editor = True
    def add_newt_widgets(self, form, current):
        label = snack.Label(self.question)
        self.button = snack.CompactButton(_("Click to edit"))
        form.add(label, form.x, form.last + 1)
        form.add(self.button, form.x, form.last + 2)
        form.last += 2

    def button_action(self, button, current):
        if type(current) != str:
            current = ''
        if falcon.conf.editor:
            editor = falcon.conf.editor
        elif os.getenv('EDITOR'):
            editor = os.getenv('EDITOR')
        elif os.path.exists('/usr/bin/editor'):
            editor = '/usr/bin/editor'
        elif os.path.exists('/usr/bin/nano'):
            nano = '/usr/bin/nano'
        elif os.path.exists('/usr/bin/vi'):
            vi = '/usr/bin/vi'
        else:
            falcon.util.error(_("Can't find an editor to use"))

        editor_blurb = "# %s\n# %s\n#\n" % (self.question, _("All lines beginning with # will be ignored"))
        (fd, path) = tempfile.mkstemp()
        try:
            fd = os.fdopen(fd,'r+')
            fd.write(editor_blurb)
            fd.write(current)
            fd.flush()
            falcon.screen.suspend()
            rc = os.spawnlp(os.P_WAIT, editor, editor, path)
            falcon.screen.resume()
            if rc:
                os.unlink(path)
                falcon.util.error(_("Launching an external editor (%s) failed, returncode: %d") % (editor, code))
            fd.seek(0)
            value = ''.join([x for x in fd.readlines() if not x.startswith('#')])
            os.unlink(path)
        except:
            try:
                os.unlink(path)
            except:
                pass
            raise
            falcon.util.error(_("Launching an external editor (%s) failed") % editor)
        return value

class Boolean(Question):
    num_newt_widgets = 1
    def add_newt_widgets(self, form, current):
        self.cb = snack.Checkbox(self.question, current)
        form.add(self.cb, form.x, form.last+1)
        form.last += 1
    def get_answer(self):
        return self.cb.value()
        
def yesno(question, default=None):
    if falcon.conf.force_yes:
        return True
    if falcon.conf.force_no:
        return False
    #return Select(question, ((_("Yes"), True), (_("No"), False))).ask(default)
    ret = raw_input(question + " ")
    return ret.lower() in ['y','yes','ok','on','1'];

def escapeurl(url):
    url = url.replace('http://','').replace('https://','').replace('ftp://','')
    ret = ''
    for c in url:
        if c.isalnum() or c in '_%.':
            ret += c
        else:
            ret += '_'
    return ret

def humanize(size):
    suffix = ' bytes'
    if size > 1048576:
        size = round(size/1048576, 2)
        suffix = 'mB'
    elif size > 1024:
        size = round(size/1024, 2)
        suffix = 'kB'
    return '%s %s' % (size, suffix)

def download(url,cache=None,store=None):
    mtime = 0
    if cache and os.path.exists(cache):
        mtime = os.path.getmtime(cache) - time.timezone
    falcon.util.output(_("Trying to download %s") % url)
    req = urllib2.urlopen(url)
    url_mtime = time.mktime(req.headers.getdate('last-modified'))
    if cache and (url_mtime < mtime):
        del req
        return open(cache)

    # Now really download
    if cache:
        fd = open(cache,'w')
    elif store:
        fd = open(store,'w+')
    size = int(req.headers['content-length'])
    sys.stdout.write(_("   Size: %s - downloaded   0%%") % humanize(size))
    sys.stdout.flush()

    dled = oldperc = 0
    while True:
        dl = req.read(2048)
        fd.write(dl)
        dled += len(dl)
        if len(dl) == 0: break
        perc = int(float(dled)/size * 100.0)
        if perc != oldperc:
            oldperc = perc
            sys.stdout.write("\b\b\b\b%3d%%" % perc)
            sys.stdout.flush()
    print ""

    if cache:
        fd.close()
        if url.endswith('.gz'):
            os.rename(cache, cache + '.gz')
            run(['gunzip', cache + '.gz'])
        if url.endswith('.bz2'):
            os.rename(cache, cache + '.bz2')
            run(['bunzip2', cache + '.bz2'])
        return open(cache)
    fd.seek(0)
    return fd

# Validators
class Dummy: pass
validators = Dummy()
del Dummy
class ValidationError(Exception): pass

def nospaces(val):
    if re.search(r'\s', val):
        raise ValidationError(_("Found a space in '%s'") % val)
validators.nospaces = nospaces

def _hex(val):
    if not re.match(r'^[a-fA-F0-9]*$', val):
        raise ValidationError(_("'%s' is not a hexadecimal number") % val)
validators.hex = _hex

def atleastone(val):
    if not len(val):
        raise ValidationError(_("At least one item required"))
validators.atleastone = atleastone

url_re = re.compile('^https?://[a-z][-a-z0-9.+/%&?]*$', re.I)
def url(val):
    if not url_re.match(val):
        raise ValidationError(_("'%s' is not a valid url") % val)
validators.url = url

mail_re = re.compile(r'^[-a-z0-9+%._]+@[-a-z0-9.]+$', re.I)
def mailaddress(val):
    if not mail_re.match(val):
        raise ValidationError(_("'%s' is not a valid mailaddress") % val)
validators.mailaddress = mailaddress

def version(val):
    if not re.match(r'^\d+(\.\d+)*$', val):
        raise ValidationError(_("'%s' is not a valid version number")  % val)
validators.version = version

shellsafe_re = re.compile('^[-a-z0-9_+=.^%@:/ ]*$')
def shellsafe(val):
    if not shellsafe_re.match(val):
        raise ValidationError(_("'%s' contains characters that can be interpreted by the shell"))
validators.shellsafe = shellsafe

def commaseparated(func):
    def inner(val):
        for v in val.split(','):
            func(v)
    return inner
validators.commaseparated = commaseparated

def debug_exception():
     for line in traceback.format_exc().split('\n'):
        debug(line)

